@@ -28,16 +28,17 @@ from fractions import Fraction
 from hypothesis.errors import InvalidArgument, ResolutionFailed
 from hypothesis.control import assume
 from hypothesis._settings import note_deprecation
+from hypothesis.internal.cache import LRUReusedCache
 from hypothesis.searchstrategy import SearchStrategy
 from hypothesis.internal.compat import gcd, ceil, floor, hrange, \
     text_type, integer_types, get_type_hints, getfullargspec, \
     implements_iterator
 from hypothesis.internal.floats import is_negative, float_to_int, \
     int_to_float, count_between_floats
+from hypothesis.internal.coverage import check_function
 from hypothesis.internal.renaming import renamed_arguments
 from hypothesis.utils.conventions import infer, not_set
 from hypothesis.internal.reflection import proxies, required_args
-from hypothesis.internal.branchcheck import check_function
 
 __all__ = [
     'nothing',
@@ -85,9 +86,10 @@ def convert_value(v):
     return (type(v), v)
 
 
-def cacheable(fn):
-    cache = {}
+STRATEGY_CACHE = LRUReusedCache(1024)
 
+
+def cacheable(fn):
     @proxies(fn)
     def cached_strategy(*args, **kwargs):
         kwargs_cache_key = set()
@@ -97,16 +99,18 @@ def cacheable(fn):
         except TypeError:
             return fn(*args, **kwargs)
         cache_key = (
+            fn,
             tuple(map(convert_value, args)), frozenset(kwargs_cache_key))
         try:
-            return cache[cache_key]
+            return STRATEGY_CACHE[cache_key]
         except TypeError:
             return fn(*args, **kwargs)
         except KeyError:
             result = fn(*args, **kwargs)
-            cache[cache_key] = result
+            if not isinstance(result, SearchStrategy) or result.is_cacheable:
+                STRATEGY_CACHE[cache_key] = result
             return result
-    cached_strategy.__clear_cache = cache.clear
+    cached_strategy.__clear_cache = STRATEGY_CACHE.clear
     return cached_strategy
 
 
@@ -131,14 +135,15 @@ defines_strategy_with_reusable_values = base_defines_strategy(True)
 
 
 class Nothing(SearchStrategy):
-    @property
-    def is_empty(self):
+    def calc_is_empty(self, recur):
         return True
 
     def do_draw(self, data):
-        data.mark_invalid()
+        # This method should never be called because draw() will mark the
+        # data as invalid immediately because is_empty is True.
+        assert False  # pragma: no cover
 
-    def calc_has_reusable_values(self):
+    def calc_has_reusable_values(self, recur):
         return True
 
     def __repr__(self):
@@ -390,9 +395,6 @@ def tuples(*args):
     for arg in args:
         check_strategy(arg)
 
-    for arg in args:
-        if arg.is_empty:
-            return nothing()
     from hypothesis.searchstrategy.collections import TupleStrategy
     return TupleStrategy(args, tuple)
 
@@ -451,13 +453,6 @@ def lists(
         else:
             return builds(list)
     check_strategy(elements)
-    if elements.is_empty:
-        if (min_size or 0) > 0:
-            raise InvalidArgument((
-                'Cannot create non-empty lists with elements drawn from '
-                'strategy %r because it has no values.') % (elements,))
-        else:
-            return builds(list)
     if unique:
         if unique_by is not None:
             raise InvalidArgument((
@@ -490,21 +485,21 @@ def lists(
             min_size=min_size,
             key=unique_by
         )
-        return result
+    else:
+        from hypothesis.searchstrategy.collections import ListStrategy
+        if min_size is None:
+            min_size = 0
+        if average_size is None:
+            if max_size is None:
+                average_size = _AVERAGE_LIST_LENGTH
+            else:
+                average_size = (min_size + max_size) * 0.5
 
-    from hypothesis.searchstrategy.collections import ListStrategy
-    if min_size is None:
-        min_size = 0
-    if average_size is None:
-        if max_size is None:
-            average_size = _AVERAGE_LIST_LENGTH
-        else:
-            average_size = (min_size + max_size) * 0.5
-
-    return ListStrategy(
-        (elements,), average_length=average_size,
-        min_size=min_size, max_size=max_size,
-    )
+        result = ListStrategy(
+            (elements,), average_length=average_size,
+            min_size=min_size, max_size=max_size,
+        )
+    return result
 
 
 @cacheable
@@ -581,9 +576,6 @@ def fixed_dictionaries(mapping):
     check_type(dict, mapping, 'mapping')
     for v in mapping.values():
         check_strategy(v)
-    for v in mapping.values():
-        if v.is_empty:
-            return nothing()
     return FixedKeysDictStrategy(mapping)
 
 
@@ -1754,6 +1746,7 @@ def check_valid_sizes(min_size, average_size, max_size):
 _AVERAGE_LIST_LENGTH = 5.0
 
 
+@cacheable
 def deferred(definition):
     """A deferred strategy allows you to write a strategy that references other
     strategies that have not yet been defined. This allows for the easy
